/***********************************************************************
 *
 * Copyright (C) 2008-2009 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "window.h"

#include "data_model.h"
#include "date_editor.h"
#include "session_dialog.h"
#include "time_editor.h"

#include <QAction>
#include <QComboBox>
#include <QDateTime>
#include <QFile>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QInputDialog>
#include <QItemEditorFactory>
#include <QLabel>
#include <QMessageBox>
#include <QMetaProperty>
#include <QPushButton>
#include <QSettings>
#include <QTimer>
#include <QToolButton>
#include <QTreeView>
#include <QVBoxLayout>
#include <QXmlStreamReader>

/*****************************************************************************/

Window::Window(QWidget* parent)
: QWidget(parent), m_active_model(0) {
	setWindowTitle(tr("Kapow Punch Clock"));
	setWindowIcon(QIcon(":/kapow.png"));

	m_projects = new QComboBox(this);
	m_projects->setInsertPolicy(QComboBox::InsertAlphabetically);
	connect(m_projects, SIGNAL(currentIndexChanged(const QString&)), this, SLOT(projectActivated(const QString&)));

	QToolButton* add_button = new QToolButton(this);
	add_button->setText("+");
	add_button->setToolTip(tr("Add Project"));
	connect(add_button, SIGNAL(clicked()), this, SLOT(addProject()));

	m_display = new QLabel(tr("00:00:00"), this);

	m_timer = new QTimer(this);
	m_timer->setInterval(1000);
	connect(m_timer, SIGNAL(timeout()), this, SLOT(updateDisplay()));

	m_auto_timer = new QTimer(this);
	m_auto_timer->setInterval(30000);
	connect(m_auto_timer, SIGNAL(timeout()), this, SLOT(autoSave()));

	m_start = new QPushButton(tr("Start"), this);
	connect(m_start, SIGNAL(clicked()), this, SLOT(start()));

	m_stop = new QPushButton(tr("Stop"), this);
	m_stop->hide();
	connect(m_stop, SIGNAL(clicked()), this, SLOT(stop()));

	m_cancel = new QPushButton(tr("Cancel"), this);
	m_cancel->setEnabled(false);
	connect(m_cancel, SIGNAL(clicked()), this, SLOT(cancel()));

	QItemEditorFactory* factory = new QItemEditorFactory;
	factory->registerEditor(QVariant::Date, new QStandardItemEditorCreator<DateEditor>());
	factory->registerEditor(QVariant::Time, new QStandardItemEditorCreator<TimeEditor>());
	factory->registerEditor(QVariant::String, new QStandardItemEditorCreator<QLineEdit>());
	QItemEditorFactory::setDefaultFactory(factory);

	m_details = new QTreeView(this);
	m_details->setRootIsDecorated(false);
	m_details->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
	m_details->setSelectionMode(QAbstractItemView::SingleSelection);
	m_details->setSelectionBehavior(QAbstractItemView::SelectRows);
	m_details->header()->setResizeMode(QHeaderView::ResizeToContents);
	m_details->header()->setStretchLastSection(true);
	m_details->setContextMenuPolicy(Qt::ActionsContextMenu);
	connect(m_details, SIGNAL(pressed(const QModelIndex&)), this, SLOT(sessionPressed(const QModelIndex&)));

	QAction* action = new QAction(tr("Add Session"), this);
	connect(action, SIGNAL(triggered()), this, SLOT(addSession()));
	m_details->addAction(action);

	m_edit_action = new QAction(tr("Edit Session"), this);
	m_edit_action->setEnabled(false);
	connect(m_edit_action, SIGNAL(triggered()), this, SLOT(editSession()));
	m_details->addAction(m_edit_action);

	m_remove_action = new QAction(tr("Remove Session"), this);
	m_remove_action->setEnabled(false);
	connect(m_remove_action, SIGNAL(triggered()), this, SLOT(removeSession()));
	m_details->addAction(m_remove_action);

	// Load details of all projects
	DataModel* model = 0;
	QFile file("data.xml");
	if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		QXmlStreamReader xml(&file);
		while (!xml.atEnd()) {
			xml.readNext();
			if (!xml.isStartElement()) {
				continue;
			}
			if (xml.name() == QLatin1String("session")) {
				QXmlStreamAttributes attributes = xml.attributes();
				QDate date = QDate::fromString(attributes.value("date").toString(), Qt::ISODate);
				QTime start = QTime::fromString(attributes.value("start").toString(), Qt::ISODate);
				QTime stop = QTime::fromString(attributes.value("stop").toString(), Qt::ISODate);
				bool billed = attributes.value("billed").toString().toInt();
				QString note = attributes.value("note").toString();
				model->add(date, start, stop, billed, note);
			} else if (xml.name() == QLatin1String("project")) {
				model = new DataModel(this);
				QString project = xml.attributes().value("name").toString();
				m_models[project] = model;
				m_projects->blockSignals(true);
				m_projects->addItem(project);
				m_projects->blockSignals(false);
			}
		}
	}
	m_start->setEnabled(!m_models.isEmpty());

	// Load last used project
	QString project;
	if (!m_models.isEmpty()) {
		project = m_models.keys().first();
	}
	int index = m_projects->findText(QSettings().value("ActiveProject", project).toString());
	if (index > 0) {
		m_projects->setCurrentIndex(index);
	} else {
		if (!m_models.isEmpty()) {
			projectActivated(m_projects->currentText());
		}
	}

	// Append auto-saved time if it exists
	QSettings settings;
	if (settings.contains("AutoSave")) {
		if (m_active_model) {
			m_active_model->add(settings.value("AutoSave/Start").toDateTime(), settings.value("AutoSave/Stop").toDateTime());
			settings.remove("AutoSave");
		}
	}

	// Lay out window
	QHBoxLayout* horiz_layout = new QHBoxLayout;
	horiz_layout->setMargin(0);
	horiz_layout->setSpacing(6);
	horiz_layout->addWidget(m_projects, 1);
	horiz_layout->addWidget(add_button);
	horiz_layout->addSpacing(6);
	horiz_layout->addWidget(m_display);
	horiz_layout->addSpacing(6);
	horiz_layout->addWidget(m_start);
	horiz_layout->addWidget(m_stop);
	horiz_layout->addWidget(m_cancel);

	QVBoxLayout* layout = new QVBoxLayout(this);
	layout->setMargin(12);
	layout->setSpacing(12);
	layout->addLayout(horiz_layout);
	layout->addWidget(m_details);

	// Restore window geometry
	resize(750, 500);
	restoreGeometry(QSettings().value("WindowGeometry").toByteArray());
}

/*****************************************************************************/

void Window::closeEvent(QCloseEvent* event) {
	QSettings().setValue("WindowGeometry", saveGeometry());
	QWidget::closeEvent(event);
}

/*****************************************************************************/

void Window::start() {
	m_start->hide();
	m_stop->show();
	m_cancel->setEnabled(true);
	m_display->setText(tr("00:00:00"));
	m_start_time = QDateTime::currentDateTime();
	QSettings().setValue("AutoSave/Start", m_start_time);
	m_timer->start();
	m_auto_timer->start();
}

/*****************************************************************************/

void Window::stop() {
	finish();
	m_active_model->add(m_start_time, QDateTime::currentDateTime());
	save();
}

/*****************************************************************************/

void Window::cancel() {
	if (QMessageBox::question(this, tr("Question"), tr("Cancel this session?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {
		return;
	}

	finish();
	m_display->setText(tr("00:00:00"));
}

/*****************************************************************************/

void Window::updateDisplay() {
	int elapsed = QDateTime::currentDateTime().toTime_t() - m_start_time.toTime_t();
	QTime convert;
	convert = convert.addSecs(elapsed);
	m_display->setText(convert.toString("hh:mm:ss"));
}

/*****************************************************************************/

void Window::autoSave() {
	QSettings().setValue("AutoSave/Stop", QDateTime::currentDateTime());
}

/*****************************************************************************/

void Window::addProject() {
	bool ok;
	QString project;
	forever {
		project = QInputDialog::getText(this, tr("Add Project"), tr("Enter project name:"), QLineEdit::Normal, project, &ok);
		if (ok && !project.isEmpty()) {
			if (!m_models.contains(project)) {
				m_models.insert(project, new DataModel(this));
				save();
				m_start->setEnabled(true);
				m_projects->addItem(project);
				m_projects->setCurrentIndex(m_projects->findText(project));
				return;
			}
			QMessageBox::warning(this, tr("Error"), tr("A project with that name already exists."));
		} else {
			break;
		}
	}
}

/*****************************************************************************/

void Window::projectActivated(const QString& project) {
	Q_ASSERT(!project.isEmpty());
	if (m_active_model) {
		m_active_model->disconnect(SIGNAL(rowsInserted(const QModelIndex&, int, int)), this, SLOT(sessionsInserted(const QModelIndex&, int, int)));
		m_active_model->disconnect(SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)), this, SLOT(save()));
	}
	m_active_model = m_models.value(project);
	connect(m_active_model, SIGNAL(rowsInserted(const QModelIndex&, int, int)), this, SLOT(sessionsInserted(const QModelIndex&, int, int)));
	connect(m_active_model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)), this, SLOT(save()));
	m_details->setModel(m_active_model);
	m_edit_action->setEnabled(false);
	m_remove_action->setEnabled(false);
	QSettings().setValue("ActiveProject", project);
}

/*****************************************************************************/

void Window::sessionPressed(const QModelIndex& index) {
	bool enabled = !m_active_model->isBilled(index.row());
	m_edit_action->setEnabled(enabled && (index.column() < 3 || index.column() > 6));
	m_remove_action->setEnabled(enabled);
}

/*****************************************************************************/

void Window::sessionsInserted(const QModelIndex&, int, int end) {
	m_details->setCurrentIndex(m_active_model->index(end, 0));
}

/*****************************************************************************/

void Window::addSession() {
	Q_ASSERT(m_active_model != 0);
	SessionDialog dialog(this);
	forever {
		if (dialog.exec() == QDialog::Accepted) {
			if (m_active_model->add(dialog.date(), dialog.start(), dialog.stop(), false, dialog.note())) {
				save();
				return;
			}
			QMessageBox::warning(this, tr("Error"), tr("Session conflicts with other sessions."));
		} else {
			break;
		}
	}
}

/*****************************************************************************/

void Window::editSession() {
	m_details->edit(m_details->currentIndex());
}

/*****************************************************************************/

void Window::removeSession() {
	Q_ASSERT(m_active_model != 0);
	if (QMessageBox::question(this, tr("Question"), tr("Remove selected session?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes) {
		m_active_model->remove(m_details->currentIndex().row());
		save();
	}
}

/*****************************************************************************/

void Window::save() {
	QFile file("data.xml");
	file.open(QIODevice::WriteOnly | QIODevice::Text);
	QXmlStreamWriter xml(&file);
	xml.setAutoFormatting(true);
	xml.writeStartDocument();
	xml.writeStartElement("kapow");
	QMapIterator<QString, DataModel*> i(m_models);
	while (i.hasNext()) {
		i.next();
		xml.writeStartElement("project");
		xml.writeAttribute("name", i.key());
		i.value()->toXml(xml);
		xml.writeEndElement();
	}
	xml.writeEndElement();
	xml.writeEndDocument();
}

/*****************************************************************************/

void Window::finish() {
	m_timer->stop();
	m_auto_timer->stop();
	m_stop->hide();
	m_start->show();
	m_cancel->setEnabled(false);
	QSettings().remove("AutoSave");
}

/*****************************************************************************/
